from wagtail.api.v2.serializers import Field
from importlib import import_module

class StreamFieldSerializer(Field):
    def __init__(self, *args, **kwargs):
        self.serializers_instances = kwargs.pop('serializers', object())
        super(StreamFieldSerializer, self).__init__(*args, **kwargs)

    def get_serializers(self):
        return self.serializers_instances

    def to_representation(self, value):
        representation = []
        if value is None:
            return representation
        serializers = self.get_serializers()
        for child in value:
            if child.block.name not in serializers.keys():
                child_representation = child.block.get_api_representation(child.value, context=self.context)
            else:
                ser = serializers[child.block.name](context=self.context)
                child_representation = ser.to_representation(child.value)
            representation.append({
                'type': child.block.name,
                'value': child_representation,
                'id': child.id
            })
        return representation


def get_stream_field_serializers():
    # todo add serializers from django.conf.settings
    # kwargs['serializers'] = ''
    form_serializers = {
        'form_text': 'wagtail_rest_pack.generic_forms.blocks.text_block.InputBlockSerializer',
        'form_group': 'wagtail_rest_pack.generic_forms.blocks.group_block.GroupBlockSerializer',
        'form_submit': 'wagtail_rest_pack.generic_forms.blocks.submit_block.SubmitBlockSerializer',
        'form': 'wagtail_rest_pack.generic_forms.view.GetFormBuilderSerializer',
    }
    from django.apps import apps
    classes = {}
    for key,value in form_serializers.items():
        try:
            module_path, class_name = value.rsplit('.', 1)
            module = import_module(module_path)
            classes[key] = getattr(module, class_name)
        except (ImportError, AttributeError) as e:
            raise ImportError(value)
    return classes

class SettingsStreamFieldSerializer(StreamFieldSerializer):
    def __init__(self, *args, **kwargs):
        kwargs.pop('serializers', {})
        super(SettingsStreamFieldSerializer, self).__init__(*args, **kwargs)

    def get_serializers(self):
        return get_stream_field_serializers()



from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from rest_framework import serializers
from wagtail.core import blocks
from wagtail_rest_pack.generic_forms.blocks.generic_input import GenericInputBlock, visibility_choices, text_validation_choices

from django.utils.translation import gettext_lazy as _


class InputBlockSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=30)
    label = serializers.CharField(max_length=50)
    max_length = serializers.IntegerField(max_value=2000, min_value=10)
    required = serializers.ChoiceField(choices=visibility_choices)
    placeholder = serializers.CharField(max_length=80, required=False)
    validation = serializers.ChoiceField(choices=text_validation_choices)
    multiline = serializers.BooleanField()

    block_name = 'form_text'

    @staticmethod
    def block_definition():
        return InputBlockSerializer.block_name, GenericInputBlock(specific=[
            ('multiline', blocks.BooleanBlock(required=False, help_text="Víceřádkový vstup."))
        ])

    class Meta:
        fields = ('name','label','max_length','required','placeholder','validation','multiline',)

    def validate_field(self, attr):
        instance = self.context['instance']
        field_name = instance['name']
        value = attr.get(field_name, None)
        if not isinstance(value, str):
            raise ValidationError(_(f'Expected "{field_name}" to be string, but is not'))
        if instance['required'] == 'anonymous_user_only':
            if self.context['request'].user.is_authenticated:
                if value is not None:
                    raise ValidationError(_(f'Expected "{field_name}" to be not set, but it is.'))
                return {}

        if value is None:
            raise ValidationError(_(f'Expected "{field_name}" to be set, but is not'))

        max_length = instance['max_length']
        if len(value) > max_length:
            raise ValidationError(_(f'Expected "{field_name}" to be of max_length "{max_length}"'))
        if instance['validation'] == "email":
            validate_email(value)
        return {
            field_name: value
        }



from .email import EmailActionHandler

def get_handlers():
    return {
        'send_email': EmailActionHandler
    }

def handle(action, *args, **kwargs):
    value = action['value']
    assert len(value['action']) == 1, ('Only one action can be set.')
    assert len(value['response']) == 1, ('Only one response can be set.')
    response = value['response'][0]
    concrete_action = value['action'][0]
    handlers = get_handlers()
    handler = handlers.get(concrete_action['type'], None)
    assert handler is not None, ('No handler configured for %s' % concrete_action['type'])
    handler(**kwargs).handle(concrete_action['value'])
    return value['response'][0]




# Generated by Django 3.2 on 2021-05-03 18:02

from django.db import migrations
import wagtail.core.blocks
import wagtail.core.fields
import wagtail_rest_pack.generic_forms.blocks.generic_input


class Migration(migrations.Migration):

    dependencies = [
        ('generic_forms', '0003_formbuilder_content'),
    ]

    operations = [
        migrations.AlterField(
            model_name='formbuilder',
            name='content',
            field=wagtail.core.fields.StreamField([('form_text', wagtail.core.blocks.StructBlock([('name', wagtail.core.blocks.TextBlock(help_text='Název pole v ascii (např. user_email)', label='Technický název pole', max_length=30, validators=[wagtail_rest_pack.generic_forms.blocks.generic_input.is_form_field])), ('label', wagtail.core.blocks.TextBlock(label='Popisek zobrazený před inputem.', max_length=50)), ('max_length', wagtail.core.blocks.IntegerBlock(default=100, help_text='Maximální délka vstupu', label='Max', max_value=2000, min_value=10)), ('required', wagtail.core.blocks.ChoiceBlock(choices=[('always', 'Vždy'), ('anonymous_user_only', 'Pouze nepřihlášený uživatel')], label='Podmínka zobrazení pole')), ('placeholder', wagtail.core.blocks.CharBlock(help_text='Hodnota zobrazená, když nic není vyplněno.', max_length=80, required=False)), ('validation', wagtail.core.blocks.ChoiceBlock(choices=[('none', 'Žádné'), ('email', 'Email')], label='Validace')), ('multiline', wagtail.core.blocks.BooleanBlock(help_text='Víceřádkový vstup.', required=False))]))]),
        ),
    ]




from django.urls import path

from .view import FormView
from .post import PostFormView

urlpatterns = [
    path('', PostFormView.as_view(), name='form_view'),
    path('<str:name>', FormView.as_view(), name='submit_form'),
]



import re

from django.core.exceptions import ValidationError
from wagtail.core import blocks

from rest_framework import serializers
from wagtail_rest_pack.generic_forms.blocks.visibility import visibility_choices
from django.utils.translation import gettext_lazy as _


def is_form_field(value):
    if not isinstance(value, str):
        raise ValidationError(_('Only string type is allowed'))
    pattern = "^[a-zA-Z0-9_-]+$"
    pass
    if re.match(pattern, value) is None:
        raise ValidationError(_('Does not match a-z, A-Z'))

text_validation_choices = [
    ('none', _('None')),
    ('email', _('Email'))
]

class GenericInputBlock(blocks.StructBlock):
    def __init__(self, specific: [] = None, **kwargs):
        # todo validace na unikátnost name
        local_blocks = [
            ('name', blocks.TextBlock(max_length=30, help_text=_('The field name, e.g. email_field'),
                                      label=_('Technical name of a field'), validators=[is_form_field])),
            ('label', blocks.TextBlock(max_length=50, label=_('Label shown before input'))),
            ('max_length',
             blocks.IntegerBlock(label='Max', help_text=_('Max length of an input'), default=100, min_value=10,
                                 max_value=2000)),
            ('required', blocks.ChoiceBlock(choices=visibility_choices, default=visibility_choices[0][0], label=_('Condition of field visilibity'))),
            ('placeholder',
             blocks.CharBlock(max_length=80, required=False, help_text=_('Help text shown when nothing filled yet.'))),
            ('validation', blocks.ChoiceBlock(choices=text_validation_choices, default='none', label=_('Validation')))
        ]
        if specific is not None:
            local_blocks.extend(specific)
        super().__init__(local_blocks=local_blocks, **kwargs)




from django.http import HttpResponseNotFound


def local_network_only(view_func):
    def wrapped_view(*args, **kwargs):
        if args[0].headers.get('x-forwarded-for') is None:
            return view_func(*args, **kwargs)
        return HttpResponseNotFound()
    return wrapped_view



from django.db import models
from modelcluster.models import ClusterableModel
from wagtail.core.fields import StreamField
from wagtail.snippets.models import register_snippet
from wagtail_rest_pack.generic_forms.blocks.group_block import GroupBlockSerializer
from wagtail_rest_pack.generic_forms.blocks.submit_block import SubmitBlockSerializer
from wagtail_rest_pack.generic_forms.blocks.text_block import InputBlockSerializer
from wagtail_rest_pack.recaptcha.permission import AuthenticatedOrRecaptcha
from rest_framework.permissions import IsAuthenticated
from django.utils.translation import gettext as _

security_choices = {
    'recaptcha_or_user': {
        'label': _('Recaptcha or logged user'),
        'permission_classes': [AuthenticatedOrRecaptcha]
    },

    'authenticated_user_only': {
        'label': _('Authenticated user only'),
        'permission_classes': [IsAuthenticated]
    }
}


@register_snippet
class FormBuilder(ClusterableModel):
    name = models.CharField(max_length=60, validators=[], primary_key=True)
    display_name = models.TextField(max_length=100, default="")
    description = models.TextField(max_length=1000, blank=True)
    created_on = models.DateTimeField(auto_now_add=True)
    security = models.CharField(max_length=30, default=next(iter(security_choices.items()))[0],
                                choices=[(choice[0],choice[1]['label']) for choice in security_choices.items()])
    stream = StreamField(block_types=[
        InputBlockSerializer.block_definition(),
        GroupBlockSerializer.block_definition(),
        SubmitBlockSerializer.block_definition(),
    ])

    def find_action(self, action):
        submittables = [action for action in self.stream.stream_data if action['type'] == 'form_submit']
        candidates = [candidate for candidate in submittables if candidate['value']['name'] == action]
        if len(candidates) != 1:
            return None
        return candidates[0]

    def __str__(self):
        return 'Formulář "{}"'.format(self.display_name)



from django.core.exceptions import ValidationError
from rest_framework import generics
from rest_framework import serializers
from rest_framework.serializers import ModelSerializer
from wagtail_rest_pack.exception.handler import custom_exception_handler
from wagtail_rest_pack.recaptcha.permission import AuthenticatedOrRecaptcha
from wagtail_rest_pack.comments.content_type import create_content_type_id
from wagtail_rest_pack.comments.models import Comment
from django.utils.translation import gettext as _


class CreateCommentSerializer(ModelSerializer):
    body = serializers.CharField(max_length=2000, required=True)
    content_type = serializers.CharField(max_length=100)
    object_id = serializers.IntegerField()
    parent_id = serializers.PrimaryKeyRelatedField(queryset=Comment.objects.all(), source='parent', required=False)

    class Meta:
        model= Comment
        fields = ['id', 'parent_id', 'body', 'object_id', 'content_type']

    def create(self, validated_data):
        content_type = validated_data.pop('content_type')
        validated_data['content_type_id'] = create_content_type_id(validated_data['object_id'], content_type)
        parent = getattr(validated_data, 'parent', None)
        if parent is not None and parent.parent is not None:
            raise ValidationError(_('Only two level comment tree is supported.'))
        user = self.context['request'].user
        if not user.is_anonymous:
            validated_data['created_by'] = user
        return super(CreateCommentSerializer, self).create(validated_data)


class CreateCommentAPIView(generics.CreateAPIView):
    permission_classes = [AuthenticatedOrRecaptcha]
    serializer_class = CreateCommentSerializer

    def get_exception_handler(self):
        return custom_exception_handler




import json
from django.http import HttpResponseNotFound, HttpResponseBadRequest, HttpResponse
from rest_framework import serializers
from rest_framework.views import APIView
from rest_framework.permissions import AllowAny
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import BasePermission
from wagtail_rest_pack.generic_forms.models import FormBuilder

from wagtail_rest_pack.exception.handler import custom_exception_handler
from .models import FormBuilder, security_choices
from .handlers.handler import handle
from .validation import validate_form_data
from django.utils.translation import gettext as _


class PostFormSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=60)
    action = serializers.CharField(max_length=50)
    data = serializers.DictField()

    class Meta:
        fields = ['name', 'action', 'data', ]


class PostFormView(APIView):
    permission_classes = [AllowAny]
    queryset = FormBuilder.objects.all()

    def get_serializer(self, *args, **kwargs):
        return PostFormSerializer(*args, **kwargs)

    def get_exception_handler(self):
        return custom_exception_handler

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, *args, **kwargs)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data
        form: FormBuilder = self.queryset.get(name=data['name'])
        action = form.find_action(action=data['action'])
        permission_classes = security_choices[form.security]['permission_classes']
        if any([not perm().has_permission(request, self) for perm in permission_classes]):
            raise PermissionDenied()
        context = {
            'request': request,
            'form': form,
            'data': data['data']
        }
        context['validated_data'] = validate_form_data(**context)
        result = handle(action, **context)
        return HttpResponse(json.dumps({
            'stream': result
        }))



from rest_framework.exceptions import ValidationError
from rest_framework.permissions import BasePermission

from .models import get_recaptcha_instance


class AuthenticatedOrRecaptcha(BasePermission):
    def has_permission(self, request, view):
        if request.user and request.user.is_authenticated:
            return True
        try:
            get_recaptcha_instance().verify(request)
            return True
        except ValidationError:
            return False




from wagtail_rest_pack.streamfield.serializers import get_stream_field_serializers


def validate_stream_data(data, stream_data, **kwargs):
    for stream_value in stream_data:
        type = stream_value['type']
        serializer_class = kwargs['serializers'].get(type, None)
        block_value = stream_value['value']
        serializer = serializer_class(context={'instance': block_value, **kwargs})
        if hasattr(serializer, 'validate_field'):
            result = serializer.validate_field(data)
            if result is not None:
                for key, value in result.items():
                    kwargs['validated_data'][key] = value
        if block_value.get('stream',None) is not None:
            stream_data = block_value['stream']
            validate_stream_data(data, stream_data, **kwargs)
    return kwargs['validated_data']

def validate_form_data(data, **kwargs):
    kwargs['serializers'] = get_stream_field_serializers()
    kwargs['validated_data'] = {}
    return validate_stream_data(data, kwargs['form'].stream.stream_data, **kwargs)

